"""Calculate and print quantile values from a single-band raster."""

import time
import warnings

from itertools import repeat
import concurrent.futures
import multiprocessing
from math import ceil

import click
import numpy as np
import rasterio
from crick import TDigest
from scipy import stats
from scipy.stats.mstats import mquantiles
import matplotlib.pyplot as plt

import rio_terrain as rt
import rio_terrain.tools.messages as msg
from rio_terrain import __version__ as plugin_version


def tdigest_mean(digest):
    """Estimate the mean from a tdigest

    """
    ctr = digest.centroids()
    mean = np.sum(ctr['mean']*ctr['weight'])/np.sum(ctr['weight'])

    return mean


def tdigest_std(digest):
    """Estimate the standard deviation of the mean from a tdigest

    """
    ctr = digest.centroids()
    mean = tdigest_mean(digest)
    std = np.sqrt(
                (np.power((ctr['mean'] - mean), 2)*ctr['weight']).sum()
                / (ctr['weight'].sum() - 1)
                )

    return std


def tdigest_stats(digest):
    """Estimate descriptive statistics from a tdigest

    """
    minX = digest.min()
    maxX = digest.max()
    meanX = tdigest_mean(digest)
    stdX = tdigest_std(digest)

    return (minX, maxX, meanX, stdX)


def digest_window(file, window, absolute):
    '''Process worker that calculates a t-digest on a raster

    '''
    with rasterio.open(file) as src:
        img = src.read(1, window=window[1])
        img[img <= src.nodata+1] = np.nan
        arr = img[np.isfinite(img)]
        if absolute:
            arr = np.absolute(arr)
        count_ = np.count_nonzero(~np.isnan(img))
        digest_ = TDigest()
        digest_.update(arr.flatten())

    return window, digest_, count_


@click.command('quantiles', short_help="Calculate quantile values.")
@click.argument('input', nargs=1, type=click.Path(exists=True))
@click.option('-q', '--quantile', multiple=True, type=float,
              help='Print quantile value.')
@click.option('-f', '--fraction', nargs=1, default=1.0,
              help='Randomly sample a fraction of internal blocks.')
@click.option('--absolute/--no-absolute', default=False,
              help='Calculate quantiles for the absolute values.')
@click.option('--describe/--no-describe', default=False,
              help='Print descriptive statistics to the console.')
@click.option('--plot/--no-plot', default=False,
              help='Display statistics plots.')
@click.option('-j', '--jobs', 'njobs', type=int, default=multiprocessing.cpu_count(),
              help='Number of concurrent jobs to run.')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode.')
@click.version_option(version=plugin_version, message='rio-terrain v%(version)s')
@click.pass_context
def quantiles(ctx, input, quantile, fraction, absolute, describe, plot, njobs, verbose):
    """Calculate and print quantile values.

    INPUT should be a single-band raster.

    \b
    Example:
    rio quantiles elevation.tif -q 0.5 -q 0.9

    """
    if verbose:
        np.warnings.filterwarnings('default')
    else:
        np.warnings.filterwarnings('ignore')

    t0 = time.time()
    command = click.get_current_context().info_name

    count = 0

    with rasterio.open(input) as src:
        profile = src.profile
        affine = src.transform

        if njobs < 1:
            click.echo("Running quantiles in-memory")
            img = src.read(1)
            img[img <= src.nodata+1] = np.nan
            arr = img[np.isfinite(img)]
            if absolute:
                arr = np.absolute(arr)

            count = np.count_nonzero(~np.isnan(img))
            description = (arr.min(), arr.max(), arr.mean(), arr.std())
            results = zip(quantile, mquantiles(arr, np.array(quantile)))

        elif njobs == 1:
            blocks = rt.subsample(src.block_windows(), probability=fraction)
            n_blocks = ceil(rt.block_count(src.shape, src.block_shapes) * fraction)
            digest = TDigest()

            click.echo("Running quantiles with sequential t-digest")
            with click.progressbar(length=n_blocks, label='Blocks done:') as bar:
                for ij, window in blocks:
                    img = src.read(1, window=window)
                    img[img <= src.nodata+1] = np.nan
                    arr = img[np.isfinite(img[:])]
                    if absolute:
                        arr = np.absolute(arr)

                    window_count = np.count_nonzero(~np.isnan(img))
                    if window_count > 0:
                        window_digest = TDigest()
                        window_digest.update(arr.flatten())
                        digest.merge(window_digest)
                        count += window_count
                    bar.update(1)

            description = tdigest_stats(digest)
            results = zip(quantile, digest.quantile(quantile))

        else:

            blocks = rt.subsample(src.block_windows(), probability=fraction)
            n_blocks = ceil(rt.block_count(src.shape, src.block_shapes)*fraction)
            digest = TDigest()

            click.echo("Running quantiles with multiprocess t-digest")
            with concurrent.futures.ProcessPoolExecutor(max_workers=njobs) as executor, \
                    click.progressbar(length=n_blocks, label='Blocks done:') as bar:
                for (window, window_digest, window_count) in executor.map(digest_window, repeat(input), blocks, repeat(absolute)):
                    if window_count > 0:
                        digest.merge(window_digest)
                        count += window_count
                    bar.update(1)

            description = tdigest_stats(digest)
            results = zip(quantile, digest.quantile(quantile))

    click.echo((msg.COMPLETION).format(msg.printtime(t0, time.time())))

    click.echo(list(results))
    minX, maxX, meanX, stdX = description

    if describe:
        click.echo("min: {}".format(minX))
        click.echo("max: {}".format(maxX))
        click.echo("mean: {}".format(meanX))
        click.echo("std: {}".format(stdX))
        click.echo("count: {}".format(count))

    if njobs > 0 and plot is True:

        ctr = digest.centroids()

        # scaled to theoretic normal
        # calculate positions relative to standard normal distribution
        qx_predicted_norm = stats.norm.ppf(digest.cdf(ctr['mean']))
        qx_norm = np.linspace(start=stats.norm.ppf(0.001), stop=stats.norm.ppf(0.999), num=250)
        qz_norm = qx_norm*stdX + meanX
        cum_norm = stats.norm.cdf(qx_norm)

        # scaled to theoretic laplace
        # calculate positions relative to laplace distribution
        """
        qx_predicted_laplace = stats.laplace.ppf(digest.cdf(ctr['mean']))
        qx_laplace = np.linspace(start=stats.laplace.ppf(0.001), stop=stats.laplace.ppf(0.999), num=250)
        qz_laplace = qx_laplace*stdX + mean
        cum_laplace = stats.laplace.cdf(qx_laplace)
        """

        # frequency at centroid (irregular width bins)
        plt.plot(ctr['mean'], ctr['weight'], 'r.')
        plt.plot([meanX], [0], 'k+', markersize=12)
        plt.xlabel('Centroid Value')
        plt.ylabel('Counts')
        plt.title('Centroid Counts')
        plt.show()

        # histogram (equal width bins)
        nbins = 1000
        hist, bin_edges = digest.histogram(nbins)
        width = (digest.max() - digest.min())/nbins
        plt.bar(bin_edges[:-1], hist, width=width)
        plt.xlabel('Value')
        plt.ylabel('Counts')
        plt.title('Histogram')
        plt.show()

        # cumulative probability distribution
        spacing = (digest.max() - digest.min())/100
        samples = np.arange(ctr['mean'].min(), ctr['mean'].max(), spacing)
        cdf = digest.cdf(samples)

        plt.plot(samples, cdf, 'r.')
        plt.plot(qz_norm, cum_norm, linestyle='dashed', c='black')
        # plt.plot(qz_laplace, cum_laplace, linestyle='dotted', c='gray')
        plt.xlabel('Value')
        plt.ylabel('Probability')
        plt.title('Cumulative Distribution')
        plt.plot([meanX], [digest.cdf(meanX)], "k+", markersize=12)
        plt.show()

        # theoretic normal
        plt.plot(qx_predicted_norm, ctr['mean'], 'r.')
        plt.plot(qx_norm, qz_norm, linestyle='dashed', c='black')
        plt.xlabel('Standard Normal Variate')
        plt.ylabel('Value')
        plt.title('QQ-plot on theoretic standard normal')
        plt.plot([0], [meanX], "k+", markersize=12)
        plt.show()

        # theoretic laplace
        """
        plt.plot(qx_predicted_laplace, ctr['mean'], 'r.')
        plt.plot(qx_laplace, qz_laplace, linestyle='dashed', c='black')
        plt.plot([0], [mean], 'k+', markersize=12)
        plt.xlabel('laplace variate')
        plt.ylabel('Elevation (m)')
        plt.title('QQ-plot on theoretic laplace')
        plt.show()
        """
    if njobs < 1 and plot is True:

        # histogram (equal width bins)
        nbins = 100
        hist, bin_edges = np.histogram(arr, nbins)
        width = (arr.max() - arr.min())/nbins
        plt.bar(bin_edges[:-1], hist, width=width)
        plt.xlabel('Elevation (m)')
        plt.ylabel('Counts')
        plt.title('Histogram')
        plt.show()

        # cumulative probability distribution
        cdf = np.cumsum(hist)/count
        qx_predicted_norm = stats.norm.ppf(cdf)

        qx_norm = np.linspace(start=stats.norm.ppf(0.001), stop=stats.norm.ppf(0.999), num=250)
        qz_norm = qx_norm * stdX + meanX
        cum_norm = stats.norm.cdf(qx_norm)

        plt.plot(bin_edges[:-1], cdf, 'r.')
        plt.plot(qz_norm, cum_norm, linestyle='dashed', c='black')
        plt.xlabel('Value')
        plt.ylabel('Probability')
        plt.title('Cumulative Distribution')
        plt.plot([meanX], [0], "k+", markersize=12)
        plt.show()

        # theoretic normal

        # full dataset is too large!
        # zscore = (arr - meanX)/stdX
        # plt.plot(zscore, arr, 'r.')

        plt.plot(qx_predicted_norm, bin_edges[:-1], 'r.')
        plt.plot(qx_norm, qz_norm, linestyle='dashed', c='black')
        plt.xlabel('Standard Normal Variate')
        plt.ylabel('Value')
        plt.title('QQ-plot on theoretic standard normal')
        plt.plot([0], [meanX], "k+", markersize=12)
        plt.show()



"""Threshold a raster with an uncertainty raster."""

import time
import warnings
import concurrent.futures

import click
import numpy as np
import rasterio

import rio_terrain as rt
import rio_terrain.tools.messages as msg
from rio_terrain import __version__ as plugin_version


def do_threshold(img0, img1, level, default=0):
    conditions = [img0 >= img1 * level, img0 <= -img1 * level]
    choices = [1, -1]
    result = np.select(conditions, choices, default=default)
    return result


@click.command('threshold', short_help="Threshold a raster with an uncertainty raster.")
@click.argument('input', nargs=1, type=click.Path(exists=True))
@click.argument('uncertainty', nargs=1, type=click.Path(exists=True))
@click.argument('output', nargs=1, type=click.Path())
@click.argument('level', nargs=1, type=float)
@click.option('-j', '--njobs', type=int, default=1, help='Number of concurrent jobs to run.')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode.')
@click.version_option(version=plugin_version, message='rio-terrain v%(version)s')
@click.pass_context
def threshold(ctx, input, uncertainty, output, level, njobs, verbose):
    """Threshold a raster with an uncertainty raster.

    \b
    INPUT should be a single-band raster.
    UNCERTAINTY should be a single-band raster representing uncertainty.

    \b
    Example:
    rio threshold diff.tif uncertainty.tif, detected.tif 1.68

    """
    if verbose:
        np.warnings.filterwarnings('default')
    else:
        np.warnings.filterwarnings('ignore')

    t0 = time.time()
    command = click.get_current_context().info_name

    with rasterio.open(input) as src0, rasterio.open(uncertainty) as src1:

        if not rt.is_raster_intersecting(src0, src1):
            raise ValueError(msg.NONINTERSECTING)
        if not rt.is_raster_aligned(src0, src1):
            raise ValueError(msg.NONALIGNED)

        profile = src0.profile
        affine = src0.transform
        nodata = np.iinfo(np.int32).min

        if njobs >= 1:
            block_shape = (src0.block_shapes)[0]
            blockxsize = block_shape[1]
            blockysize = block_shape[0]
        else:
            blockxsize = None
            blockysize = None

        tiles = rt.tile_grid_intersection(
            src0, src1, blockxsize=blockxsize, blockysize=blockysize
        )
        windows0, windows1, write_windows, affine, nrows, ncols = tiles

        profile.update(
            dtype=rasterio.int32,
            nodata=nodata,
            count=1,
            height=nrows,
            width=ncols,
            transform=affine,
            compress='lzw',
            bigtiff='yes',
        )

        with rasterio.open(output, 'w', **profile) as dst:
            if njobs < 1:
                click.echo((msg.STARTING).format(command, msg.INMEMORY))
                img0 = src0.read(1, window=next(windows0))
                img1 = src1.read(1, window=next(windows1))
                result = do_threshold(img0, img1, level, default=nodata)
                dst.write(result, 1)
            elif njobs == 1:
                click.echo((msg.STARTING).format(command, msg.SEQUENTIAL))
                with click.progressbar(length=nrows * ncols, label='Blocks done:') as bar:
                    for (window0, window1, write_window) in zip(windows0, windows1, write_windows):
                        img0 = src0.read(1, window=window0)
                        img1 = src1.read(1, window=window1)
                        result = do_threshold(img0, img1, level, default=nodata)
                        dst.write(result, 1, window=write_window)
                        bar.update(result.size)
            else:
                click.echo((msg.STARTING).format(command, msg.CONCURRENT))

                def jobs():
                    for (window0, window1, write_window) in zip(windows0, windows1, write_windows):
                        img0 = src0.read(1, window=window0)
                        img1 = src1.read(1, window=window1)
                        yield img0, img1, window0, window1, write_window

                with concurrent.futures.ThreadPoolExecutor(max_workers=njobs) as executor, \
                        click.progressbar(length=nrows * ncols, label='Blocks done:') as bar:

                    future_to_window = {
                        executor.submit(
                            do_threshold, img0, img1, level, default=nodata
                        ): (window0, window1, write_window)
                        for (img0, img1, window0, window1, write_window) in jobs()
                    }

                    for future in concurrent.futures.as_completed(future_to_window):
                        window0, window1, write_window = future_to_window[future]
                        result = future.result()
                        dst.write(result, 1, window=write_window)
                        bar.update(result.size)

    click.echo((msg.WRITEOUT).format(output))
    click.echo((msg.COMPLETION).format(msg.printtime(t0, time.time())))



"""Extract regions from a raster by a data range."""

import time
import warnings
import concurrent.futures

import click
import numpy as np
import rasterio

import rio_terrain as rt
import rio_terrain.tools.messages as msg
from rio_terrain import __version__ as plugin_version


def do_slice(img, minimum=None, maximum=None, keep_data=False, false_val=0):
    """Slice data or ones from an array given a value range.

    Parameters:
        img (ndarray)
        minimum (float)
        maximum (float)
        keep_data (bool)

    Returns:
        result (ndarray)

    """

    # default bounds
    if minimum is None:
        minimum = np.nanmin(img)
    if maximum is None:
        maximum = np.nanmax(img)

    if keep_data:
        result = np.where((img >= minimum) & (img <= maximum), img, false_val)
    else:
        result = np.where((img >= minimum) & (img <= maximum), 1, false_val)

    return result


@click.command('slice', short_help="Extract regions by data range.")
@click.argument('input', nargs=1, type=click.Path(exists=True))
@click.argument('output', nargs=1, type=click.Path())
@click.option('--minimum', nargs=1, type=float, default=None, help='Minimum value to extract.')
@click.option('--maximum', nargs=1, type=float, default=None, help='Maximum value to extract.')
@click.option('--keep-data/--no-keep-data', is_flag=True,
              help='Return the input data, or return ones.')
@click.option('--zeros/--no-zeros', is_flag=True,
              help='Use the raster nodata value or zeros for False condition.')
@click.option('-j', '--njobs', type=int, default=1, help='Number of concurrent jobs to run.')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode.')
@click.version_option(version=plugin_version, message='rio-terrain v%(version)s')
@click.pass_context
def slice(ctx, input, output, minimum, maximum, keep_data, zeros, njobs, verbose):
    """Extract regions from a raster by a data range.

    INPUT should be a single-band raster.

    \b
    Setting the --keep-data option will return the data values.
    The default is to return a raster of ones and zeros.

    \b
    Example:
    rio range diff.tif extracted.tif --minumum -2.0 --maximum 2.0

    """
    if verbose:
        np.warnings.filterwarnings('default')
    else:
        np.warnings.filterwarnings('ignore')

    t0 = time.time()
    command = click.get_current_context().info_name

    with rasterio.open(input) as src:

        profile = src.profile
        affine = src.transform

        if keep_data:
            dtype = profile['dtype']
            nodata = profile['nodata']
            profile.update(count=1, compress='lzw')
        else:
            dtype = 'int32'
            nodata = np.iinfo(np.int32).min
            profile.update(
                dtype=rasterio.int32, nodata=nodata, count=1, compress='lzw'
            )

        if zeros:
            false_val = 0
        else:
            false_val = nodata

        with rasterio.open(output, 'w', **profile) as dst:
            if njobs < 1:
                click.echo((msg.STARTING).format(command, msg.INMEMORY))
                img = src.read(1)
                result = do_slice(img, minimum, maximum, keep_data, false_val)
                dst.write(result.astype(dtype), 1)
            elif njobs == 1:
                click.echo((msg.STARTING).format(command, msg.SEQUENTIAL))
                with click.progressbar(length=src.width * src.height, label='Blocks done:') as bar:
                    for (ij, window) in src.block_windows():
                        img = src.read(1, window=window)
                        result = do_slice(img, minimum, maximum, keep_data, false_val)
                        dst.write(result.astype(dtype), 1, window=window)
                        bar.update(result.size)
            else:
                click.echo((msg.STARTING).format(command, msg.CONCURRENT))

                def jobs():
                    for (ij, window) in src.block_windows():
                        img = src.read(1, window=window)
                        yield img, window

                with concurrent.futures.ThreadPoolExecutor(max_workers=njobs) as executor, \
                        click.progressbar(length=src.width*src.height, label='Blocks done:') as bar:

                    future_to_window = {
                        executor.submit(
                            do_slice, img, minimum, maximum, keep_data, false_val
                        ): (window)
                        for (img, window) in jobs()
                    }

                    for future in concurrent.futures.as_completed(future_to_window):
                        window = future_to_window[future]
                        result = future.result()
                        dst.write(result.astype(dtype), 1, window=window)
                        bar.update(result.size)

    click.echo((msg.WRITEOUT).format(output))
    click.echo((msg.COMPLETION).format(msg.printtime(t0, time.time())))



"""Subtract one single-band raster from another."""

import time
import warnings
import concurrent.futures

import numpy as np
import click
import rasterio

import rio_terrain as rt
import rio_terrain.tools.messages as msg
from rio_terrain import __version__ as plugin_version


@click.command('difference', short_help="Subtract one raster from another.")
@click.argument('input_t0', nargs=1, type=click.Path(exists=True))
@click.argument('input_t1', nargs=1, type=click.Path(exists=True))
@click.argument('output', nargs=1, type=click.Path())
@click.option('-b', '--blocks', 'blocks', nargs=1, type=int, default=40,
              help='Multiple internal blocks to chunk.')
@click.option('-j', '--njobs', type=int, default=1, help='Number of concurrent jobs to run.')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode.')
@click.version_option(version=plugin_version, message='rio-terrain v%(version)s')
@click.pass_context
def difference(ctx, input_t0, input_t1, output, blocks, njobs, verbose):
    """Subtract INPUT_T0 from INPUT_T1.

    \b
    INPUT_T0 should be a single-band raster at time t0.
    INPUT_T1 should be a single-band raster at time t1.

    \b
    Example:
    rio diff elevation1.tif elevation2.tif, diff2_1.tif

    """
    if verbose:
        np.warnings.filterwarnings('default')
    else:
        np.warnings.filterwarnings('ignore')

    t0 = time.time()
    command = click.get_current_context().info_name

    with rasterio.open(input_t0) as src0, rasterio.open(input_t1) as src1:

        if not rt.is_raster_intersecting(src0, src1):
            raise ValueError(msg.NONINTERSECTING)
        # if not rt.is_raster_aligned(src0, src1):
        #    raise ValueError(msg.NONALIGNED)

        profile = src0.profile
        affine = src0.transform

        if njobs >= 1:
            block_shape = (src0.block_shapes)[0]
            blockxsize = block_shape[1]
            blockysize = block_shape[0]
        else:
            blockxsize = None
            blockysize = None

        tiles = rt.tile_grid_intersection(
            src0, src1, blockxsize=blockxsize * blocks, blockysize=blockysize * blocks)
        windows0, windows1, write_windows, affine, nrows, ncols = tiles

        profile.update(
            dtype=rasterio.float32,
            count=1,
            height=nrows,
            width=ncols,
            transform=affine,
            compress='lzw',
            bigtiff='yes',
        )

        with rasterio.open(output, 'w', **profile) as dst:
            if njobs < 1:
                click.echo((msg.STARTING).format(command, msg.INMEMORY))
                img0 = src0.read(1, window=next(windows0))
                img1 = src1.read(1, window=next(windows1))
                img0[img0 <= src0.nodata + 1] = np.nan
                img1[img1 <= src1.nodata + 1] = np.nan
                result = img1 - img0
                dst.write(result.astype(profile['dtype']), 1, window=next(write_windows))
            elif njobs == 1:
                click.echo((msg.STARTING).format(command, msg.SEQUENTIAL))
                with click.progressbar(length=nrows * ncols, label='Blocks done:') as bar:
                    for (window0, window1, write_window) in zip(windows0, windows1, write_windows):
                        img0 = src0.read(1, window=window0)
                        img1 = src1.read(1, window=window1)
                        img0[img0 <= src0.nodata + 1] = np.nan
                        img1[img1 <= src1.nodata + 1] = np.nan
                        result = img1 - img0
                        dst.write(result.astype(profile['dtype']), 1, window=write_window)
                        bar.update(result.size)
            else:
                click.echo((msg.STARTING).format(command, msg.CONCURRENT))

                def jobs():
                    for (window0, window1, write_window) in zip(windows0, windows1, write_windows):
                        img0 = src0.read(1, window=window0)
                        img1 = src1.read(1, window=window1)
                        img0[img0 <= src0.nodata + 1] = np.nan
                        img1[img1 <= src1.nodata + 1] = np.nan
                        yield img0, img1, window0, window1, write_window

                def diff(img0, img1):
                    return img1 - img0

                with concurrent.futures.ThreadPoolExecutor(max_workers=njobs) as executor, \
                        click.progressbar(length=nrows * ncols, label='Blocks done:') as bar:

                    future_to_window = {
                        executor.submit(diff, img0, img1): (
                            window0,
                            window1,
                            write_window,
                        )
                        for (img0, img1, window0, window1, write_window) in jobs()
                    }

                    for future in concurrent.futures.as_completed(future_to_window):
                        window0, window1, write_window = future_to_window[future]
                        result = future.result()
                        dst.write(result.astype(profile['dtype']), 1, window=write_window)
                        bar.update(result.size)

    click.echo((msg.WRITEOUT).format(output))
    click.echo((msg.COMPLETION).format(msg.printtime(t0, time.time())))



"""Extract regions from a raster by category."""

import time
import warnings
import concurrent.futures

import click
import numpy as np
import rasterio

import rio_terrain as rt
import rio_terrain.tools.messages as msg
from rio_terrain import __version__ as plugin_version


def do_extract(img, categorical, category):
    if category is None:
        _category = [1]
    else:
        _category = list(category)
    mask = np.isin(categorical, _category)
    result = img * mask
    return result


@click.command('extract', short_help="Extract regions by category.")
@click.argument('input', nargs=1, type=click.Path(exists=True))
@click.argument('categorical', nargs=1, type=click.Path(exists=True))
@click.argument('output', nargs=1, type=click.Path())
@click.option('-c', '--category', multiple=True, type=int, help='Category to extract.')
@click.option('-j', '--njobs', type=int, default=1, help='Number of concurrent jobs to run')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode.')
@click.version_option(version=plugin_version, message='rio-terrain v%(version)s')
@click.pass_context
def extract(ctx, input, categorical, output, category, njobs, verbose):
    """Extract regions from a raster by category.

    \b
    INPUT should be a single-band raster.
    CATEGORICAL should be a single-band raster with categories to extract.

    The categorical data may be the input raster or another raster.

    \b
    Example:
    rio extract diff.tif categorical.tif extract.tif -c 1 -c 3

    """
    if verbose:
        np.warnings.filterwarnings('default')
    else:
        np.warnings.filterwarnings('ignore')

    t0 = time.time()
    command = click.get_current_context().info_name

    with rasterio.open(input) as src, rasterio.open(categorical) as cat:

        if not rt.is_raster_intersecting(src, cat):
            raise ValueError(msg.NONINTERSECTING)
        if not rt.is_raster_aligned(src, cat):
            raise ValueError(msg.NONALIGNED)

        profile = src.profile
        affine = src.transform

        if njobs >= 1:
            block_shape = (src.block_shapes)[0]
            blockxsize = block_shape[1]
            blockysize = block_shape[0]
        else:
            blockxsize = None
            blockysize = None

        tiles = rt.tile_grid_intersection(src, cat, blockxsize=blockxsize, blockysize=blockysize)
        windows0, windows1, write_windows, affine, nrows, ncols = tiles

        profile.update(
            count=1,
            compress='lzw',
            bigtiff='yes',
            height=nrows,
            width=ncols,
            transform=affine,
        )

        with rasterio.open(output, 'w', **profile) as dst:
            if njobs < 1:
                click.echo((msg.STARTING).format(command, msg.INMEMORY))
                img = src.read(1, window=next(windows0))
                mask = cat.read(1, window=next(windows1))
                result = do_extract(img, mask, category)
                dst.write(result.astype(profile['dtype']), 1, window=next(write_windows))
            elif njobs == 1:
                click.echo((msg.STARTING).format(command, msg.SEQUENTIAL))
                with click.progressbar(length=nrows * ncols, label='Blocks done:') as bar:
                    for (window0, window1, write_window) in zip(windows0, windows1, write_windows):
                        img = src.read(1, window=window0)
                        mask = cat.read(1, window=window1)
                        result = do_extract(img, mask, category)
                        dst.write(result.astype(profile['dtype']), 1, window=write_window)
                        bar.update(result.size)
            else:
                click.echo((msg.STARTING).format(command, msg.CONCURRENT))

                def jobs():
                    for (window0, window1, write_window) in zip(windows0, windows1, write_windows):
                        img = src.read(1, window=window0)
                        mask = cat.read(1, window=window1)
                        yield img, mask, window0, window1, write_window

                with concurrent.futures.ThreadPoolExecutor(max_workers=njobs) as executor, \
                        click.progressbar(length=nrows * ncols, label='Blocks done:') as bar:

                    future_to_window = {
                        executor.submit(do_extract, img, mask, category): (
                            window0,
                            window1,
                            write_window,
                        )
                        for (img, mask, window0, window1, write_window) in jobs()
                    }

                    for future in concurrent.futures.as_completed(future_to_window):
                        window0, window1, write_window = future_to_window[future]
                        result = future.result()
                        dst.write(result.astype(profile['dtype']), 1, window=write_window)
                        bar.update(result.size)

    click.echo((msg.WRITEOUT).format(output))
    click.echo((msg.COMPLETION).format(msg.printtime(t0, time.time())))



"""Calculate a standard-deviation raster."""

import time
import warnings
import concurrent.futures
from math import ceil

import click
import numpy as np
import rasterio

import rio_terrain as rt
import rio_terrain.tools.messages as msg
from rio_terrain.core import focalstatistics
from rio_terrain import __version__ as plugin_version


@click.command('std', short_help="Calculate standard-deviation.")
@click.argument('input', nargs=1, type=click.Path(exists=True))
@click.argument('output', nargs=1, type=click.Path())
@click.option('-n', '--neighborhood', nargs=1, default=3, help='Neigborhood size in cells.')
@click.option('-b', '--blocks', 'blocks', nargs=1, type=int, default=40,
              help='Multiple internal blocks to chunk.')
@click.option('-j', '--njobs', type=int, default=1, help='Number of concurrent jobs to run.')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode.')
@click.version_option(version=plugin_version, message='rio-terrain v%(version)s')
@click.pass_context
def std(ctx, input, output, neighborhood, blocks, njobs, verbose):
    """Calculate a standard-deviation raster.

    INPUT should be a single-band raster.

    \b
    Example:
    rio std elevation.tif stddev.tif

    """
    if verbose:
        np.warnings.filterwarnings('default')
    else:
        np.warnings.filterwarnings('ignore')

    t0 = time.time()
    command = click.get_current_context().info_name

    with rasterio.open(input) as src:
        profile = src.profile
        affine = src.transform
        profile.update(dtype=rasterio.float32, count=1, compress='lzw', bigtiff='yes')

        if (njobs >= 1) and src.is_tiled:
            blockshape = (list(src.block_shapes))[0]
            if (blockshape[0] == 1) or (blockshape[1] == 1):
                warnings.warn((msg.STRIPED).format(blockshape))
            read_windows = rt.tile_grid(
                src.width,
                src.height,
                blockshape[0] * blocks,
                blockshape[1] * blocks,
                overlap=neighborhood,
            )
            write_windows = rt.tile_grid(
                src.width,
                src.height,
                blockshape[0] * blocks,
                blockshape[1] * blocks,
                overlap=0,
            )
        else:
            blockshape = 128
            warnings.warn((msg.NOTILING).format(src.shape))

        with rasterio.open(output, 'w', **profile) as dst:
            if njobs < 1:
                click.echo((msg.STARTING).format(command, msg.INMEMORY))
                img = src.read(1)
                img[img <= src.nodata + 1] = np.nan
                result = focalstatistics.std(img, size=(neighborhood, neighborhood))
                dst.write(result.astype(profile['dtype']), 1)
            elif njobs == 1:
                click.echo((msg.STARTING).format(command, msg.SEQUENTIAL))
                with click.progressbar(length=src.width * src.height, label='Blocks done:') as bar:
                    for (read_window, write_window) in zip(read_windows, write_windows):
                        img = src.read(1, window=read_window)
                        img[img <= src.nodata + 1] = np.nan
                        arr = focalstatistics.std(img, size=(neighborhood, neighborhood))
                        result = rt.trim(arr, rt.margins(read_window, write_window))
                        dst.write(result.astype(profile['dtype']), 1, window=write_window)
                        bar.update(result.size)
            else:
                click.echo((msg.STARTING).format(command, msg.CONCURRENT))
                import dask.array as da

                with click.progressbar(length=src.width * src.height, label='Blocks done:') as bar:
                    for (read_window, write_window) in zip(read_windows, write_windows):
                        img = src.read(1, window=read_window)
                        img[img <= src.nodata + 1] = np.nan

                        chunks_wanted = 100
                        bw = ceil(read_window.width / np.sqrt(chunks_wanted) / blockshape[0])
                        bh = ceil(read_window.height / np.sqrt(chunks_wanted) / blockshape[1])
                        hh, ww = rt.chunk_dims(
                            (img.shape[0], img.shape[1]),
                            (blockshape[0] * bw, blockshape[1] * bh),
                            min_size=blockshape[0] * 2)

                        arr = da.from_array(img, chunks=(tuple(hh), tuple(ww)))
                        tiles_in = da.overlap.overlap(
                            arr,
                            depth={0: neighborhood, 1: neighborhood},
                            boundary={0: np.nan, 1: np.nan})
                        tiles_out = tiles_in.map_blocks(
                            focalstatistics.std,
                            size=(neighborhood, neighborhood),
                            dtype=np.float64)
                        trim_out = da.overlap.trim_internal(
                            tiles_out, {0: neighborhood, 1: neighborhood})
                        full_result = trim_out.compute()
                        result = rt.trim(full_result, rt.margins(read_window, write_window))

                        dst.write(result.astype(profile['dtype']), 1, window=write_window)
                        bar.update(result.size)

                """
                def jobs():
                    for (read_window, write_window) in zip(read_windows, write_windows):
                        img = src.read(1, window=read_window)
                        img[img <= src.nodata+1] = np.nan
                        yield img, read_window, write_window

                with concurrent.futures.ThreadPoolExecutor(max_workers=njobs) as executor, \
                        click.progressbar(length=src.width*src.height, label='Blocks done:') as bar:

                    future_to_window = {
                        executor.submit(
                            focalstatistics.std,
                            img,
                            size=(neighborhood, neighborhood)): (read_window, write_window)
                        for (img, read_window, write_window) in jobs()}

                    for future in concurrent.futures.as_completed(future_to_window):
                        read_window, write_window = future_to_window[future]
                        arr = future.result()
                        result = rt.trim(arr, rt.margins(read_window, write_window))
                        dst.write(result, 1, window=write_window)
                        bar.update(result.size)
                """

    click.echo((msg.WRITEOUT).format(output))
    click.echo((msg.COMPLETION).format(msg.printtime(t0, time.time())))



"""Label regions in a single-band raster."""

import time
import warnings
import concurrent.futures

import click
import numpy as np
import rasterio
import scipy

import rio_terrain as rt
import rio_terrain.tools.messages as msg
from rio_terrain import __version__ as plugin_version


BOX = np.ones((3, 3))
CROSS = np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0]])


@click.command('label', short_help="Label regions.")
@click.argument('input', nargs=1, type=click.Path(exists=True))
@click.argument('output', nargs=1, type=click.Path())
@click.option('--diagonals/--no-diagonals', 'diagonals', default=False,
              help='Connect diagonal cells (8-connected).')
@click.option('--zeros/--no-zeros', is_flag=True,
              help='Use the raster nodata value or zeros for False condition.')
@click.option('-j', '--njobs', type=int, default=0, help='Number of concurrent jobs to run.')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode.')
@click.version_option(version=plugin_version, message='rio-terrain v%(version)s')
@click.pass_context
def label(ctx, input, output, diagonals, zeros, njobs, verbose):
    """Label regions in a raster.

    INPUT should be a single-band raster.
    
    \b
    Example:
    rio label blobs.tif labeled_blobs.tif

    """
    if verbose:
        np.warnings.filterwarnings('default')
    else:
        np.warnings.filterwarnings('ignore')

    t0 = time.time()
    command = click.get_current_context().info_name

    with rasterio.open(input) as src:

        profile = src.profile
        affine = src.transform

        dtype = 'int32'
        nodata = np.iinfo(np.int32).min
        profile.update(dtype=rasterio.int32, nodata=nodata, count=1, compress='lzw')

        if zeros:
            false_val = 0
        else:
            false_val = nodata

        if diagonals is True:
            structure = BOX
        else:
            structure = CROSS

        with rasterio.open(output, 'w', **profile) as dst:
            if njobs < 1:
                click.echo((msg.STARTING).format(command, msg.INMEMORY))
                img = src.read(1)
                labels, count = scipy.ndimage.label(img, structure=structure)
                labels[img == nodata] = false_val
                labels[labels == 0] = false_val
                dst.write(labels.astype(dtype), 1)
            elif njobs == 1:
                click.echo((msg.STARTING).format(command, msg.SEQUENTIAL))

                blockshape = (list(src.block_shapes))[0]
                if (blockshape[0] == 1) or (blockshape[1] == 1):
                    warnings.warn((msg.STRIPED).format(blockshape))
                read_windows = rt.tile_grid(
                    src.width, src.height, src.width, blockshape[1], overlap=1)
                write_windows = rt.tile_grid(
                    src.width, src.height, src.width, blockshape[1], overlap=0)

                overlap = None
                total = 0
                for (read_window, write_window) in zip(read_windows, write_windows):
                    img = src.read(1, window=read_window)
                    labels, count = scipy.ndimage.label(img, structure=structure)

                    if overlap is not None:
                        # remap values based on overlap
                        stack = np.stack([np.squeeze(overlap), np.squeeze(img[:1, :])])
                        pairs = np.unique(stack.T, axis=0)
                        idx = np.all(pairs > 0, axis=1)
                        print(pairs[idx])

                        # now remap
                        # can i use numpy.place and extract here?
                        # this strategy won't work in one pass - consider a meander that crosses back and forth over three chunks
                    else:
                        print('overlap empty')
                    total += count
                    # this will need to be the overlap after remapping labels
                    overlap = labels[-1:, :]

                    print(total)
                    # result = labels
                    # dst.write(result.astype(dtype), 1, window=write_window)
            else:
                click.echo((msg.STARTING).format(command, msg.CONCURRENT))
                click.echo('NOT IMPLEMENTED')

    click.echo('Wrote {} labels to {}'.format(count, output))
    click.echo((msg.COMPLETION).format(msg.printtime(t0, time.time())))



import sys

# Startup information
STARTING = "Running {} on {}"
INMEMORY = "in-memory raster"
SEQUENTIAL = "sequential raster blocks"
CONCURRENT = "concurrent raster blocks"

# Completion status
COMPLETION = "Finished in {}"
WRITEOUT = "Wrote output to {}"

# Warnings
STRIPED = "Blocks are lines with shape {}. Rewrite the data blocks for sequential and parallel processing."
NOTILING = "Raster with shape {} is not tiled. Rewrite the data with tiling for sequential and parallel processing."

# Errors
NONINTERSECTING = "Input rasters are non-intersecting"
NONALIGNED = "Raster cells are not aligned between inputs"


def printtime(t0, t1):
    """Return the elapsed time between t0 and t1 in h:m:s formatted string

    """
    m, s = divmod(t1 - t0, 60)
    h, m = divmod(m, 60)
    fmt = '%d:%02d:%02d' % (h, m, s)

    return fmt


def progress(value, endvalue, bar_length=20, msg=None):
    """Display and update a progress bar

    """
    done_char = '#'
    todo_char = ' '

    percent = float(value) / endvalue
    progress = done_char * int(round(percent * bar_length))
    spaces = todo_char * (bar_length - len(progress))

    message = "\rPercent: [{0}] {1}%    {2}".format(
        progress + spaces, int(round(percent * 100)), msg
    )
    sys.stdout.write(message)
    if value == endvalue - 1:
        sys.stdout.write('\n')
    sys.stdout.flush()
